#!/usr/bin/env python

"""
OpenID provider login resources which redirect clients back to the application
("relying party").

Copyright (C) 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

import WebStack.Generic
from WebStack.Helpers.Auth import Authenticator, check_openid_signature, make_openid_signature
import datetime
import time
import random
import cgi # for escape

class OpenIDLoginUtils:

    "Utilities for OpenID login screens which may be inherited."

    openid_ns = "http://specs.openid.net/auth/2.0"
    signed_names = ["op_endpoint", "return_to", "response_nonce", "assoc_handle", "claimed_id", "identity"]

    def __init__(self, associations=None, use_redirect=1):
        self.associations = associations or {}
        self.use_redirect = use_redirect

    def urlencode(self, trans, value):
        if not trans.default_charset:
            return trans.encode_path(value, self.urlencoding)
        else:
            return trans.encode_path(value)

    def get_openid_fields(self, trans, claimed_id, local_id, username, return_to, endpoint):

        # Make an association that can be used in signature verification.
        # NOTE: Probably need to consider the secret key a bit more.

        handle = username + str(time.time())
        secret_key = str(random.randint(0, 999999999))
        self.associations[handle] = secret_key

        # Make a timestamp.

        now = datetime.datetime.utcnow()
        timestamp = now.strftime("%Y-%m-%dT%H:%M:%SZ") + str(now.microsecond)

        # Make a signature.

        fields = {
            "openid.ns" : [self.openid_ns],
            "openid.mode" : ["id_res"],
            "openid.signed" : [",".join(self.signed_names)],
            "openid.op_endpoint" : [endpoint],
            "openid.return_to" : [return_to],
            "openid.response_nonce" : [timestamp],
            "openid.assoc_handle" : [handle],
            "openid.claimed_id" : [claimed_id],
            "openid.identity" : [local_id]
            }
        signature = make_openid_signature(self.signed_names, fields, secret_key)
        fields["openid.sig"] = [signature]

        return fields

    def get_openid_url(self, trans, fields):

        # Build an URL for returning to the application.

        url = "%s?" % fields["openid.return_to"][0]

        first = 1
        for name, value in fields.items():
            if not first:
                url += "&"
            url += "%s=%s" % (name, self.urlencode(trans, value[0]))
            first = 0

        return url

    def redirect_to_application(self, trans, claimed_id, local_id, username, return_to, endpoint):

        """
        Redirect the client using 'trans', 'claimed_id', 'local_id', 'username'
        and the given 'return_to' and 'endpoint' details.
        """

        fields = self.get_openid_fields(trans, claimed_id, local_id, username, return_to, endpoint)
        url = self.get_openid_url(trans, fields)

        # Show the success page anyway.
        # Offer a POST-based form for redirection.

        self.show_success(trans, fields)
        if self.use_redirect:
            trans.redirect(url)
        else:
            raise WebStack.Generic.EndOfResponse

    def show_verification(self, trans, status):

        """
        Writes a signature verification response using the transaction 'trans'
        and the 'status' of the verification.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/plain"))
        out = trans.get_response_stream()

        # NOTE: Need to use invalidate_handle, too.

        if status:
            status_str = "true"
        else:
            status_str = "false"
        out.write("ns:%s\nis_valid:%s\n" % (self.openid_ns, status_str))
        raise WebStack.Generic.EndOfResponse

    def check_authentication(self, trans, fields):

        "Check the authentication details supplied in 'trans' and 'fields'."

        # Obtain the secret key from recorded associations.

        handle = fields.get("openid.assoc_handle", [None])[0]
        if handle is not None and self.associations.has_key(handle):
            valid = check_openid_signature(fields, self.associations[handle])
            del self.associations[handle]
        else:
            valid = 0

        # Produce a response for this request.

        self.show_verification(trans, valid)

    def check_login(self, trans, fields):

        "Check the login details supplied in 'trans' and 'fields'."

        return_to = fields.get("openid.return_to", [""])[0]
        claimed_id = fields.get("openid.claimed_id", [""])[0]
        local_id = fields.get("openid.identity", [""])[0]

        # Check a combination of local identifier and username together with
        # the password.

        username = fields.get("username", [""])[0]
        password = fields.get("password", [""])[0]

        # NOTE: Permit flexibility in the credentials.

        if self.authenticator.authenticate(trans, (local_id, username), password):
            endpoint = self.app_url + trans.get_path_without_query(self.urlencoding)
            self.redirect_to_application(trans, claimed_id, local_id, username, return_to, endpoint)

class OpenIDLoginResource(OpenIDLoginUtils):

    "A resource providing a login screen."

    def __init__(self, app_url, authenticator, associations=None, use_redirect=1, urlencoding=None, encoding=None):

        """
        Initialise the resource with the application URL 'app_url' and an
        'authenticator'.

        The optional 'associations' is a mapping from association handles to
        secret keys.

        If the optional 'use_redirect' flag is set to a false value (which is
        not the default), a confirmation screen is given instead of immediately
        redirecting the user back to the original application.

        The optional 'urlencoding' parameter allows a special encoding to be
        used in producing the redirection path.

        The optional 'encoding' parameter allows a special encoding to be used
        in producing the login pages.

        To change the pages employed by this resource, either redefine the
        'login_page' and 'success_page' attributes in instances of this class or
        a subclass, or override the 'show_login' and 'show_success' methods.
        """

        OpenIDLoginUtils.__init__(self, associations, use_redirect)
        self.app_url = app_url
        self.authenticator = authenticator
        self.urlencoding = urlencoding
        self.encoding = encoding

    def respond(self, trans):

        "Respond using the transaction 'trans'."

        # Check for a submitted login form.

        fields = trans.get_fields(self.encoding)

        if fields.has_key("login"):
            self.check_login(trans, fields)
            # The above method may not return.

        # Check for an OpenID signature verification request.

        elif fields.get("openid.mode", [None])[0] == "check_authentication":
            self.check_authentication(trans, fields)
            # The above method does not return.

        # NOTE: Permit association requests here.
        # Otherwise, show the login form.

        self.show_login(trans, fields)

    def show_login(self, trans, fields):

        """
        Writes a login screen using the transaction 'trans' and 'fields'.
        """

        return_to = fields.get("openid.return_to", [""])[0]
        claimed_id = fields.get("openid.claimed_id", [""])[0]
        local_id = fields.get("openid.identity", [""])[0]

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding))
        out = trans.get_response_stream()
        out.write(self.login_page % tuple(map(cgi.escape, (return_to, claimed_id, local_id))))

    def show_success(self, trans, fields):

        """
        Writes a success screen using the transaction 'trans', using a
        dictionary of 'fields' providing details of the transaction.
        """

        trans.set_content_type(WebStack.Generic.ContentType("text/html", self.encoding))
        out = trans.get_response_stream()
        l = []
        for name, values in fields.items():
            l.append("""<input name="%s" type="hidden" value="%s" />""" % (name, cgi.escape(values[0])))
        out.write(self.success_page % (fields["openid.return_to"][0], "\n".join(l)))

    login_page = """
<html>
  <head>
    <title>Login</title>
  </head>
  <body>
    <h1>Login</h1>
    <form method="POST">
      <p>Username: <input name="username" type="text" size="12"/></p>
      <p>Password: <input name="password" type="password" size="12"/></p>
      <p><input name="login" type="submit" value="Login"/></p>
      <input name="openid.return_to" type="hidden" value="%s"/>
      <input name="openid.claimed_id" type="hidden" value="%s"/>
      <input name="openid.identity" type="hidden" value="%s"/>
    </form>
  </body>
</html>
"""

    success_page = """
<html>
  <head>
    <title>Login Example</title>
  </head>
  <body>
    <h1>Login Successful</h1>
    <form action="%s" method="POST" name="openid_redirect">
      %s
      <p>Please proceed to the application: <input name="proceed" type="submit" value="Proceed!" /></p>
    </form>
  </body>
</html>
"""

# vim: tabstop=4 expandtab shiftwidth=4
